#include <stdio.h>
#include <stdlib.h>
#include "gb_flip.h"

/*
  Defines
*/
#define MAX 5001  /* Poderemos tratar rvores com no mximo 2500 ns
		     internos */

/*
  Constantes
*/
const char no_externo[] = "\\Tn";
const char no_interno[] = "\\Tc*{2pt}";
const char tree_sep[] = "5mm";
const char level_sep[] = "5mm";

/*
  Variveis globais
*/
char arvore[MAX];  /* Seqncia de bits que representa uma rvore
		      binria */
unsigned int ipl_acumul;  /* Para o clculo do IPL mdio */
unsigned int altura_acumul;  /* Para o clculo da altura mdia */

/*
  Prottipos
*/
int calcula_ipl_altura(int i, int nivel, int n);
int faz_desenho(int i, int nivel, int n);

/*
  Funo main
*/
int main()
{
  int i, j, k;
  int n;  /* Nmero de ns internos */
  int s;  /* Semente do gerador de nmeros aleatrios */
  int r;  /* Nmero de repeties do experimento */
  int desenhar = 0;  /* Indica se  necessrio fazer o desenho */
  
  /* L os dados de entrada */
  fprintf(stderr, "* Digite quantos ns internos deve ter a rvore binria (n <= 2500): ");
  scanf("%d", &n);
  fprintf(stderr, "* Entre com uma semente para o gerador de nmeros aleatrios: ");
  scanf("%d", &s);
  fprintf(stderr, "* Nmero de repeties do experimento (digite -1 para desenhar uma rvore): ");
  scanf("%d", &r);

  gb_init_rand(s);

  /* Executa o experimento ou faz o desenho */
  for (i = 0; i < n; i++) arvore[i] = 1;
  for ( ; i < 2 * n + 1; i++) arvore[i] = 0;

  ipl_acumul = altura_acumul = 0;

  if (r == -1) { 
    r = 1;
    desenhar = 1;
  }

  for (i = 0; i < r; i++) {
    int soma;

    /* Permuta o vetor rvore ao acaso */
    for (j = 0; j < 2 * n + 1; j++) {
      char t;
      int p = j + gb_unif_rand(2 * n + 1 - j);
      
      t = arvore[p];
      arvore[p] = arvore[j];
      arvore[j] = t;
    }

    /* Verifica a partir de qual posio  uma rvore */
    for (j = 0; ; ) {
      if (arvore[j] == 0) {
	j++;
	continue;
      }

      soma = 1; k = (j + 1) % (2 * n + 1);
      while (k != j && soma >= 0) {
	if (arvore[k] == 1) soma += 1;
	else soma -= 1;

	k = (k + 1) % (2 * n + 1);
      }

      if (k == j) break;
      j = k;
    }

    /* Calcula o ipl e a altura */
    calcula_ipl_altura(j, 0, n);
  }

  if (!desenhar)
    printf("IPL mdio: %f\n", (double) ipl_acumul / r);
  else {
    faz_desenho(j, 0, n);
    fprintf(stderr, "* IPL: %d\n", ipl_acumul);
  }

  return 0;
}

/*
  calcula_ipl_altura
  Esta funo calcula o IPL e a altura de uma rvore representada por
  uma seqncia de bits que a codifica como se tivesse sido percorrida
  em pr-ordem. A seqncia deve se encontrar na varivel global
  arvore.

  A funo recebe o ndice a partir do qual deve iniciar a
  interpretao da seqncia, o nvel do n que est analisando no
  momento e o nmero de ns internos da rvore inicial. Ela devolve o
  nmero de bits que foram lidos da seqncia original. A funo
  tambm utiliza a varivel global altura para determinar a altura da
  rvore. Ela deve ser ZERO antes da primeira chamada.
*/
int calcula_ipl_altura(int i, int nivel, int n)
{
  int d1, d2;

  if (arvore[i] == 0) return 1;  /* Caso base */

  ipl_acumul += nivel;
  d1 = calcula_ipl_altura((i + 1) % (2 * n + 1), nivel + 1, n);
  d2 = calcula_ipl_altura((i + 1 + d1) % (2 * n + 1), nivel + 1, n);

  return (d1 + d2 + 1);
}

/*
  faz_desenho
  Percorre a rvore representada pela seqncia de bits dada pelo
  vetor global arvore a partir de uma posio j (consideramos a
  seqncia circular) e envia para a sada padro um cdigo para o
  pacote PStree que contm o desenho da rvore.

  A funo recebe a posio j a partir da qual comea a ver a
  seqncia, o nvel que est observando (inicialmente zero) e o
  nmero de ns da rvore que a seqncia representa. Ela devolve o
  nmero de bits que leu da seqncia.

  OBS.: Altere as variveis globais ...
*/
int faz_desenho(int i, int nivel, int n)
{
  int d1, d2;

  if (arvore[i] == 0) {
    printf("%s\n", no_externo);
    return 1;
  }

  if (nivel == 0)
    printf("\\pstree[levelsep=%s, treesep=%s]{%s}{\n", level_sep,
	   tree_sep, no_interno);
  else
    printf("\\pstree{%s}{\n", no_interno);
  d1 = faz_desenho((i + 1) % (2 * n + 1), nivel + 1, n);
  d2 = faz_desenho((i + 1 + d1) % (2 * n + 1), nivel + 1, n);
  printf("}\n");
  return d1 + d2 + 1;
}
